package org.artifactory.metadata;

import org.artifactory.event.work.NodeEventsErrorsBackOffStrategy;
import org.artifactory.storage.db.event.error.model.NodeEventsErrorInfo;

import static java.util.concurrent.TimeUnit.HOURS;

/**
 * @author Mor Merhav
 */
public class MetadataBackOffStrategy implements NodeEventsErrorsBackOffStrategy {

    private static final long MAX_HOURS_TO_WAIT = HOURS.toMillis(30);

    /**
     * Indicate whether error is being retried too frequently and should have backoff delay before next
     * retry attempt. Days elapsed since the error initial creation is compared to a time defined by the backoff
     * strategy.
     * Backoff strategy:
     * First attempt (retryCount == 0) is always allowed. From the 2nd attempt onward the number of hours to wait
     * before retrying is equal to the retry count * 6, up to a max of 30 hours
     */
    @Override
    public boolean isBackoffDelayRequired(NodeEventsErrorInfo error, long currentTime) {
        int errorCount = error.getErrorCount();
        if (errorCount == 0) {
            return false;
        }
        long errorCreatedTime = error.getFirstErrorTime();
        long millisSinceErrorCreated = currentTime - errorCreatedTime;
        return millisSinceErrorCreated < backoff(errorCount);
    }

    /**
     * retryCount   #  0   1   2    3    4    5  ...
     * time (Hours) #  0   6   12   24   30   30 ...
     */
    private long backoff(int retryCount) {
        long daysToWaitBeforeNextErrorRetry = HOURS.toMillis(retryCount * 6L);
        return Math.min(daysToWaitBeforeNextErrorRetry, MAX_HOURS_TO_WAIT);
    }

}
