/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2020 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.artifactory.metrics.framework.providers;

import org.artifactory.repo.http.HttpConnectionPoolMonitorService;
import org.artifactory.repo.http.mbean.HTTPConnectionPool;
import org.jetbrains.annotations.NotNull;
import org.jfrog.metrics.model.Metric;
import org.jfrog.metrics.model.MetricProvider;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.artifactory.metrics.framework.providers.util.ProviderUtil.addMetricWithLabelsToList;

/**
 * Provider to collect remote repository http connection metrics
 *
 * @author leela
 */
public class HttpConnectionsMetricProvider implements MetricProvider {
    private static final String METRIC_AVAILABLE = "available";
    private static final String METRIC_LEASED = "leased";
    private static final String METRIC_PENDING = "pending";
    private static final String METRIC_MAX = "max";
    private final HttpConnectionPoolMonitorService httpConnectionPoolMonitorService;


    public HttpConnectionsMetricProvider(HttpConnectionPoolMonitorService httpConnectionPoolMonitorService) {
        this.httpConnectionPoolMonitorService = httpConnectionPoolMonitorService;
    }

    private static boolean filterUnnecessaryPools(
            Map.Entry<String, WeakReference<HTTPConnectionPool>> stringWeakReferenceEntry) {
        WeakReference<HTTPConnectionPool> weakReference = stringWeakReferenceEntry.getValue();
        if (weakReference != null) {
            HTTPConnectionPool pool = weakReference.get();
            return pool != null && pool.getAvailable() > 0;
        }
        return false;
    }

    private static int comparebyPoolAvailable(Map.Entry<String, WeakReference<HTTPConnectionPool>> e1,
            Map.Entry<String, WeakReference<HTTPConnectionPool>> e2) {
        if (e2.getValue() != null && e1.getValue() != null) {
            HTTPConnectionPool pool2 = e2.getValue().get();
            HTTPConnectionPool pool1 = e1.getValue().get();
            if (pool2 != null && pool1 != null) {
                return pool2.getAvailable() - pool1.getAvailable();
            }
        }
        return 0;
    }


    @Override
    public String getName() {
        return "jfrt_http_connections";
    }

    @Override
    public String getDescription() {
        return "Artifactory HTTP Connection Metrics Provider";
    }

    @Override
    public long getRefreshIntervalMillis() {
        return 0;
    }

    @Override
    public List<Metric> collect() {
        List<Metric> metrics = new ArrayList<>();
        if (httpConnectionPoolMonitorService != null) {
            Map<String, WeakReference<HTTPConnectionPool>> activeHttpPools =
                    getActiveHttpPools(httpConnectionPoolMonitorService.getConnectionPools());
            if (!activeHttpPools.isEmpty()) {
                activeHttpPools.forEach((poolName, pool) -> {
                    Map<String, Object> labels = new HashMap<>();
                    getListOfMetricsToRegister(metrics, poolName, pool, labels);
                });
            }
        }
        return metrics;
    }

    private void getListOfMetricsToRegister(List<Metric> metrics, String poolName,
            WeakReference<HTTPConnectionPool> poolWeakReference, Map<String, Object> labels) {
        HTTPConnectionPool pool = poolWeakReference.get();
        if (pool != null) {
            int available = pool.getAvailable();
            int max = pool.getMax();
            int leased = pool.getLeased();
            int pending = pool.getPending();
            labels.put("pool", poolName);
            labels.put(METRIC_MAX, String.valueOf(max));
            addMetricWithLabelsToList(metrics, labels, METRIC_AVAILABLE, Metric.MetricType.COUNTER,
                    "Available Connections", available);
            addMetricWithLabelsToList(metrics, labels, METRIC_LEASED, Metric.MetricType.COUNTER,
                    "Leased Connections", leased);
            addMetricWithLabelsToList(metrics, labels, METRIC_PENDING, Metric.MetricType.COUNTER,
                    "Pending Connections", pending);
            addMetricWithLabelsToList(metrics, labels, METRIC_MAX, Metric.MetricType.COUNTER,
                    "Max Connections", max);
        }
    }

    @NotNull
    private Map<String, WeakReference<HTTPConnectionPool>> getActiveHttpPools(
            Map<String, WeakReference<HTTPConnectionPool>> httpPools) {
        return httpPools.entrySet()
                .stream()
                .filter(HttpConnectionsMetricProvider::filterUnnecessaryPools)
                .sorted(HttpConnectionsMetricProvider::comparebyPoolAvailable)
                .limit(10)
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
    }

    @Override
    public ReportType getReportType() {
        return ReportType.CONTINUOUS;
    }

    @Override
    public long getLastUpdatedMillis() {
        return 0L;
    }
}
