package org.artifactory.repo.interceptor;

import lombok.AccessLevel;
import lombok.Getter;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.build.BuildAddon;
import org.artifactory.api.build.BuildWrapper;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.build.BuildServiceUtils;
import org.artifactory.build.InternalBuildService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.MutableStatusHolder;
import org.artifactory.exception.CancelException;
import org.artifactory.fs.ItemInfo;
import org.artifactory.md.Properties;
import org.artifactory.repo.RepoPath;
import org.artifactory.sapi.fs.VfsItem;
import org.artifactory.sapi.interceptor.ImportInterceptor;
import org.artifactory.sapi.interceptor.context.DeleteContext;
import org.artifactory.sapi.interceptor.context.InterceptorCreateContext;
import org.artifactory.sapi.interceptor.context.InterceptorMoveCopyContext;
import org.artifactory.storage.fs.MutableVfsFile;
import org.jfrog.build.api.Build;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;

import javax.annotation.Nullable;

import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.apache.http.HttpStatus.SC_CONFLICT;
import static org.apache.http.HttpStatus.SC_FORBIDDEN;
import static org.artifactory.build.BuildServiceUtils.buildUnderWrongPath;

/**
 * Intercept actions performed on Build Info repository
 *
 * @author Yuval Reches
 */
@Getter(AccessLevel.PROTECTED)
public class BuildInfoInterceptor extends AbstractBuildInfoInterceptor implements ImportInterceptor {
    private static final Logger log = LoggerFactory.getLogger(BuildInfoInterceptor.class);

    private static final String ERR_INNER_COPY_MOVE =
            "Copy and Move operations are not allowed within the Build Info repository";
    private static final String ERR_BAD_PATH =
            "Trying to position build '%s:%s' under wrong path '%s' is not allowed, " +
                    "the expected path for this build-info json is '%s'";
    private final AuthorizationService authService;
    private final AddonsManager addonsManager;

    @Autowired
    public BuildInfoInterceptor(RepositoryService repositoryService, InternalBuildService buildService,
            AuthorizationService authService, AddonsManager addonsManager) {
        super(repositoryService, buildService);
        this.authService = authService;
        this.addonsManager = addonsManager;
    }

    @Override
    public void beforeCreate(VfsItem<? extends ItemInfo> fsItem, MutableStatusHolder statusHolder) {
        RepoPath path = fsItem.getRepoPath();
        if (isActionWithNonJsonFileInBuildInfoRepo(fsItem, path)) {
            //Reporting the CancelException in the status holder will fail the deployment.
            String err = "The '" + path.getRepoKey() + "' repository rejected the deployment of '" +
                    path.getPath() + "'. Only valid Build Info .json files are supported.";
            statusHolder.error(err, SC_CONFLICT, new CancelException(err, SC_CONFLICT), log);
        }
    }

    @Override
    public void afterCreate(VfsItem<? extends ItemInfo> fsItem, MutableStatusHolder statusHolder,
            InterceptorCreateContext ctx) {
        if (shouldTakeAction(fsItem)) {
            doAction(fsItem, statusHolder, ctx);
        }
    }

    @Override
    public void afterImport(VfsItem<? extends ItemInfo> fsItem, MutableStatusHolder statusHolder) {
        if (shouldTakeAction(fsItem)) {
            doAction(fsItem, statusHolder, null);
        }
    }

    private void doAction(VfsItem<? extends ItemInfo> fsItem, MutableStatusHolder statusHolder,
            @Nullable InterceptorCreateContext ctx) {
        RepoPath deploymentPath = fsItem.getRepoPath();
        String fullPath = deploymentPath.toPath();
        log.debug("Build Info json deployed to '{}'", fullPath);
        BuildWrapper buildWrapper = buildService.getBuildModelFromFile(deploymentPath);
        Build build = buildWrapper.getBuild();
        //This is the path the json should actually go to
        RepoPath targetPath = BuildServiceUtils.generateBuildJsonRepoPath(build, buildWrapper.getBuildRepo());
        // We don't process builds under wrong path, it will be deleted by caller
        if (buildUnderWrongPath(build, deploymentPath, buildWrapper.getBuildRepo())) {
            log.debug("Build deployed to wrong path, ignoring it");
            return;
        }
        if (fsItem instanceof MutableVfsFile && !((MutableVfsFile) fsItem).isOriginallyNew() ||
                (ConstantValues.buildsDbForceOverrideEnabled.getBoolean() &&
                        buildService.isBuildExistsInDb(buildWrapper))) {
            //If this is an override we just need to kick the build out of the db before re-adding it.
            log.info("Override of build at path '{}' detected.", fullPath);
            log.debug("Removing build that matches path '{}' from db", fullPath);
            buildService.deleteBuildInternal(build, buildWrapper.getBuildRepo(), statusHolder);
        }
        // In case of error - exception is thrown and deployment of .json file fails
        buildService.addBuildInternal(buildWrapper);
        // Update the response with the right coordinates of the artifact after it has been moved
        if (ctx != null) {
            ctx.setAlternateRepoPath(targetPath);
        }
    }

    /**
     * In case {@link DeleteContext#isAvoidBuildDeleteInterceptor()} is on, it means we wish to delete the entire
     * repo and not trigger build delete per build number / build name.
     *
     * Upon delete of a folder this interceptor is called upon every file inside.
     */
    @Override
    public void afterDelete(VfsItem<? extends ItemInfo> fsItem, MutableStatusHolder statusHolder, DeleteContext ctx) {
        if (!ctx.isAvoidBuildDeleteInterceptor() && shouldTakeAction(fsItem)) {
            RepoPath buildJsonPath = fsItem.getRepoPath();
            if (log.isDebugEnabled()) {
                log.debug("Trying to delete build according to file '{}'", fsItem.getRepoPath().toPath());
            }
            BuildWrapper buildFromFile = buildService.getBuildModelFromFile(buildJsonPath);
            // Validating path here in order to assume the path in the buildService later
            if (buildUnderWrongPath(buildFromFile.getBuild(), buildJsonPath, buildFromFile.getBuildRepo())) {
                if (log.isDebugEnabled()) {
                    log.debug("Build {}:{} is deleted from repo only, since its path doesn't match the layout: {}",
                            buildFromFile.getBuild().getName(), buildFromFile.getBuild().getNumber(),
                            buildJsonPath.toPath());
                }
                return;
            }
            //By default does not remove artifacts (since interceptor might also be triggered by undeploy)
            // REST calls do the build artifact deletion manually
            buildService.deleteBuildInternal(buildFromFile.getBuild(), buildFromFile.getBuildRepo(), statusHolder);
        }
    }

    /**
     * We allow move/copy into the repo to support copy/move from REST and trashcan restore (which is move).
     * Copy and Move within the repo makes no sense since the deployment path is calculated based on the content of the
     * json and obviously since the content doesn't change when you copy/move then the json will end in the same path.
     */
    @Override
    public boolean isCopyOrMoveAllowed(VfsItem<? extends ItemInfo> sourceItem, RepoPath targetPath,
            MutableStatusHolder status) {
        if (shouldTakeAction(sourceItem, targetPath)) {
            RepoPath sourcePath = sourceItem.getRepoPath();
            if (isBuildRepo(sourcePath) && !isBuildRepo(targetPath)) {
                status.error(ERR_INNER_COPY_MOVE, SC_BAD_REQUEST,
                        new CancelException(ERR_INNER_COPY_MOVE, SC_BAD_REQUEST), log);
                return false;
            } else {
                try {
                    BuildWrapper buildToAdd = buildService.getBuildModelFromFile(sourcePath);
                    RepoPath expectedPath =
                            BuildServiceUtils.generateBuildJsonRepoPath(buildToAdd.getBuild(), targetPath.getRepoKey());
                    if (!expectedPath.equals(targetPath)) {
                        String err = badPathError(buildToAdd.getBuild(), expectedPath, targetPath);
                        status.error(err, SC_BAD_REQUEST, new CancelException(err, SC_BAD_REQUEST), log);
                        return false;
                    }
                } catch (Exception e) {
                    log.debug("", e);
                    String err =
                            "Failed to parse Build Info from source path '" + sourcePath + "'. Operation is blocked: " +
                                    e.getMessage();
                    status.error(err, SC_BAD_REQUEST, new CancelException(err, SC_BAD_REQUEST), log);
                    return false;
                }
            }
        }
        return true;
    }

    @Override
    public void beforeCopy(VfsItem<? extends ItemInfo> sourceItem, RepoPath targetRepoPath,
            MutableStatusHolder statusHolder, Properties properties) {
        isCopyOrMoveAllowed(sourceItem, targetRepoPath, statusHolder);
    }

    @Override
    public void beforeMove(VfsItem<? extends ItemInfo> sourceItem, RepoPath targetRepoPath,
            MutableStatusHolder statusHolder, Properties properties) {
        isCopyOrMoveAllowed(sourceItem, targetRepoPath, statusHolder);
    }

    @Override
    public void afterMove(VfsItem<? extends ItemInfo> sourceItem, VfsItem<? extends ItemInfo> targetItem,
            MutableStatusHolder statusHolder, Properties properties, InterceptorMoveCopyContext ctx) {
        afterMoveCopy(sourceItem, targetItem, statusHolder);
    }

    @Override
    public void afterCopy(VfsItem<? extends ItemInfo> sourceItem, VfsItem<? extends ItemInfo> targetItem,
            MutableStatusHolder statusHolder, Properties properties, InterceptorMoveCopyContext ctx) {
        afterMoveCopy(sourceItem, targetItem, statusHolder);
    }

    /**
     * {@link #isCopyOrMoveAllowed} limits the scope of this method to run only on copy/move operations *into* the
     * build info repo.
     */
    private void afterMoveCopy(VfsItem<? extends ItemInfo> sourceItem, VfsItem<? extends ItemInfo> targetItem,
            MutableStatusHolder status) {
        if (shouldTakeAction(targetItem)) {
            BuildWrapper buildToAdd = buildService.getBuildModelFromFile(targetItem.getRepoPath());
            RepoPath expectedPath =
                    BuildServiceUtils.generateBuildJsonRepoPath(buildToAdd.getBuild(), targetItem.getRepoKey());
            RepoPath targetPath = targetItem.getRepoPath();
            if (!expectedPath.equals(targetPath)) {
                //Shouldn't really reach here since the 'before' and 'allowed' triggers prevent this but what the heck
                String err = badPathError(buildToAdd.getBuild(), expectedPath, targetPath);
                status.error(err, SC_BAD_REQUEST, new CancelException(err, SC_BAD_REQUEST), log);
                return;
            }

            buildService.addBuildInternal(buildToAdd);
            addonsManager.addonByType(BuildAddon.class)
                    .setBuildRepoPropertyOnBuilds(buildToAdd.getBuild(), sourceItem.getRepoKey(),
                            targetItem.getRepoKey());
        }
    }

    @Override
    public void assertDeleteRepoAllowed(String repoKey, MutableStatusHolder status) {
        String defaultBuildRepo = buildService.getBuildInfoRepoForDefaultProject();
        if (defaultBuildRepo.equals(repoKey)) {
            status.error("Deletion of Build Info repository " + repoKey + " is not allowed.", SC_FORBIDDEN, log);
        } else if (buildService.getAllBuildInfoRepos().contains(repoKey) && !authService.isAdmin()) {
            status.error("Deletion of Build-Info repository '" + repoKey + "' is restricted only for super admin.",
                    SC_FORBIDDEN, log);
        }
    }


    private boolean isActionWithNonJsonFileInBuildInfoRepo(VfsItem<? extends ItemInfo> vfsItem, RepoPath targetPath) {
        return vfsItem.isFile() && !isJsonFile(targetPath) && isBuildRepo(targetPath);
    }

    private String badPathError(Build build, RepoPath expectedPath, RepoPath targetPath) {
        return String
                .format(ERR_BAD_PATH, build.getName(), build.getNumber(), targetPath.toPath(), expectedPath.toPath());
    }
}