package org.artifactory.download;

import org.apache.commons.compress.archivers.ArchiveEntry;
import org.apache.commons.compress.archivers.ArchiveOutputStream;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.input.CountingInputStream;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.io.checksum.policy.LocalRepoChecksumPolicy;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.repo.RepoPath;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.traffic.TrafficService;
import org.artifactory.traffic.v2.TrafficEntryV2;
import org.jfrog.common.archive.ArchiveType;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.Charset;
import java.util.Optional;

import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

public class FolderArchiveStreamerTrafficTest extends ArtifactoryHomeBoundTest {

    private static final String DATA = "test";
    private static final int SIZE = 8;

    private ArtifactoryContext context;

    private ArchiveEntry archiveEntry;
    private ArchiveOutputStream archiveOutputStream;
    private RepoPath repoPath;
    private Optional<CountingInputStream> countingInputStream;
    private TrafficService trafficService;

    @BeforeMethod
    public void setUp() {
        context = Mockito.mock(ArtifactoryContext.class);
        when(context.beanForType(AuthorizationService.class)).thenReturn(mock(AuthorizationService.class));
        when(context.beanForType(RepositoryService.class)).thenReturn(mock(RepositoryService.class));
        ArtifactoryContextThreadBinder.bind(context);
        trafficService = Mockito.mock(TrafficService.class);
        when(context.beanForType(TrafficService.class)).thenReturn(trafficService);
        repoPath = new RepoPathImpl("a", "b");
        archiveEntry = mock(ArchiveEntry.class);
        archiveOutputStream = mock(ArchiveOutputStream.class);
        InputStream stream = IOUtils.toInputStream(DATA, Charset.defaultCharset());
        countingInputStream = Optional.of(new CountingInputStream(stream));
    }

    @AfterMethod
    public void afterMethod() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void happyFlow() throws IOException {
        FolderArchiveStreamer streamer = new FolderArchiveStreamer(repoPath, ArchiveType.TGZ,
                1, 1, new LocalRepoChecksumPolicy(), true, true);
        streamer.writeArchiveEntry(archiveEntry, archiveOutputStream, SIZE, repoPath, countingInputStream);

        TrafficEntryV2 captured = getCapturedEntity(trafficService);
        assertValid(captured, SIZE, DATA.length());
    }

    @Test(expectedExceptions = IOException.class)
    public void exceptionWhileProcessing() throws IOException {
        doThrow(new IOException()).when(archiveOutputStream).putArchiveEntry(any());

        FolderArchiveStreamer streamer = new FolderArchiveStreamer(repoPath, ArchiveType.TGZ,
                1, 1, new LocalRepoChecksumPolicy(), true, true);
        streamer.writeArchiveEntry(archiveEntry, archiveOutputStream, SIZE, repoPath, countingInputStream);

        TrafficEntryV2 captured = getCapturedEntity(trafficService);
        assertValid(captured, SIZE, DATA.length());
    }

    private TrafficEntryV2 getCapturedEntity(TrafficService trafficService) {
        ArgumentCaptor<TrafficEntryV2> trafficEntryV2Captor = ArgumentCaptor.forClass(TrafficEntryV2.class);
        verify(trafficService, times(1)).handleTrafficEntry(trafficEntryV2Captor.capture());
        TrafficEntryV2 captured = trafficEntryV2Captor.getValue();
        return captured;
    }

    private void assertValid(TrafficEntryV2 captured, long fileSize, long transferred) {
        assertEquals(captured.getAction(), TrafficEntryV2.TrafficAction.DOWNLOAD);
        assertEquals(captured.getActor(), TrafficEntryV2.TrafficActor.USER);
        assertEquals(captured.getRedirectTarget(), TrafficEntryV2.TrafficTarget.DIRECT);
        assertEquals(captured.getRepoPath(), "a:b");
        assertEquals(captured.getSize(), fileSize);
        assertEquals(captured.getRequestedSize(), fileSize);
        assertEquals(captured.getTransferredSize(), transferred);
    }

}