package org.artifactory.engine;

import org.apache.commons.io.input.CountingInputStream;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.request.ArtifactoryResponse;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.ConstantValues;
import org.artifactory.fs.RepoResource;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.request.ArtifactoryRequest;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.traffic.TrafficService;
import org.artifactory.traffic.v2.TrafficEntryV2;
import org.mockito.ArgumentCaptor;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.io.IOException;
import java.io.InputStream;
import java.util.function.Function;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.fail;

public class UploadServiceTrafficLogTest extends ArtifactoryHomeBoundTest {

    ArtifactoryContext context;

    @BeforeClass
    public void init() {
        context = mock(ArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterClass
    public void unbind() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void withoutError() {
        TrafficEntryV2 captured = getTrafficEntry(false);
        assertValid(captured);
    }

    @Test
    public void withError() {
        TrafficEntryV2 captured = getTrafficEntry(true);
        assertValid(captured);
    }

    private void assertValid(TrafficEntryV2 captured) {
        assertEquals(captured.getAction(), TrafficEntryV2.TrafficAction.UPLOAD);
        assertEquals(captured.getActor(), TrafficEntryV2.TrafficActor.USER);
        assertEquals(captured.getRedirectTarget(), TrafficEntryV2.TrafficTarget.DIRECT);
        assertEquals(captured.getUserAddress(), "127.0.0.1");
        assertEquals(captured.getRepoPath(), "a:b");
        assertEquals(captured.getSize(), 42);
        assertEquals(captured.getRequestedSize(), 42);
        assertEquals(captured.getTransferredSize(), 40);
    }

    private TrafficEntryV2 getTrafficEntry(boolean simulateError) {
        AuthorizationService authorizationService = mock(AuthorizationService.class);
        when(authorizationService.currentUsername()).thenReturn("");
        when(context.beanForType(AuthorizationService.class)).thenReturn(authorizationService);
        homeStub.getArtifactoryProperties()
                .setProperty(ConstantValues.test.getPropertyName(), "true");
        TrafficService trafficService = mock(TrafficService.class);
        CountingInputStream countingInputStream = mock(CountingInputStream.class);
        when(countingInputStream.getByteCount()).thenReturn(40L);
        UploadServiceTestStub uploadService = new UploadServiceTestStub(trafficService,
                inputStream -> countingInputStream);
        ArtifactoryRequest request = mock(ArtifactoryRequest.class);
        ArtifactoryResponse response = mock(ArtifactoryResponse.class);
        LocalRepo repo = mock(LocalRepo.class);
        RepoResource res = mock(RepoResource.class);
        when(res.getSize()).thenReturn(42L);
        RepoPath repoPath = new RepoPathImpl("a", "b");
        when(res.getRepoPath()).thenReturn(repoPath);
        uploadService.simulateError = simulateError;
        try {
            uploadService.uploadItemWithProvidedContent(request, response, repo, res);
        } catch (Exception e) {
            if (!simulateError) {
                fail();
            }
        }
        ArgumentCaptor<TrafficEntryV2> captor = ArgumentCaptor.forClass(TrafficEntryV2.class);
        verify(trafficService, times(1))
                .handleTrafficEntry(captor.capture());
        return captor.getValue();
    }

    private static class UploadServiceTestStub extends UploadServiceImpl {

        private boolean simulateError;

        public UploadServiceTestStub(TrafficService trafficService,
                Function<InputStream, CountingInputStream> streamFactory) {
            super();
            setTrafficService(trafficService);
            setStreamFactory(streamFactory);
        }

        @Override
        void uploadItemWithContent(ArtifactoryRequest request, ArtifactoryResponse response, LocalRepo repo,
                RepoResource res, InputStream inputStream) throws IOException {
            if (simulateError) {
                throw new IOException();
            }
        }
    }
}
