package org.artifactory.livelogs;

import com.google.common.collect.Lists;
import edu.emory.mathcs.backport.java.util.Collections;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ArtifactoryRunningMode;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.state.ArtifactoryServerState;
import org.artifactory.storage.db.servers.model.ArtifactoryServer;
import org.artifactory.storage.db.servers.model.ArtifactoryServerRole;
import org.artifactory.storage.db.servers.service.ArtifactoryServersCommonService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.io.File;
import java.util.List;

import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertThrows;

public class LiveLogsServiceImplTest extends ArtifactoryHomeBoundTest {

    public static final String ARTIFACTORY_SERVICE_LOG = "artifactory-service.log";
    public static final String ARTIFACTORY_REQUEST_LOG = "artifactory-request.log";
    public static final String ARTIFACTORY_TRAFFIC_LOG = "artifactory-traffic-v2.log";
    public static final String NODE_ID_1 = "nodeId1";
    public static final String NODE_ID_2 = "nodeId2";

    @Mock
    private ArtifactoryServersCommonService artifactoryServersCommonService;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private AddonsManager addonsManager;

    private LiveLogsService liveLogsService;

    @BeforeMethod
    public void beforeMethod() {
        MockitoAnnotations.openMocks(this);
        File logDir = Mockito.mock(File.class);
        ArtifactoryHome home = Mockito.mock(ArtifactoryHome.class);
        ArtifactoryHome.bind(home);
        when(home.getLogDir()).thenReturn(logDir);

        liveLogsService =
                spy(new LiveLogsServiceImpl(artifactoryServersCommonService, authorizationService, addonsManager));
        File artifactoryServiceLog = Mockito.mock(File.class);
        when(artifactoryServiceLog.getName()).thenReturn(ARTIFACTORY_SERVICE_LOG);
        when(artifactoryServiceLog.length()).thenReturn(10L);
        when(artifactoryServiceLog.isFile()).thenReturn(true);

        File artifactoryRequestLog = Mockito.mock(File.class);
        when(artifactoryRequestLog.getName()).thenReturn(ARTIFACTORY_REQUEST_LOG);
        when(artifactoryRequestLog.length()).thenReturn(10L);
        when(artifactoryRequestLog.isFile()).thenReturn(true);

        File artifactoryTrafficLog = Mockito.mock(File.class);
        when(artifactoryTrafficLog.getName()).thenReturn(ARTIFACTORY_TRAFFIC_LOG);
        when(artifactoryTrafficLog.length()).thenReturn(10L);
        when(artifactoryTrafficLog.isFile()).thenReturn(true);

        when(logDir.listFiles())
                .thenReturn(new File[] {artifactoryServiceLog, artifactoryRequestLog, artifactoryTrafficLog});

        ArtifactoryServer rt1 =
                new ArtifactoryServer(NODE_ID_1, 1612800485851L, "URL1", 1, ArtifactoryServerState.RUNNING,
                        ArtifactoryServerRole.MEMBER, 1612800485851L, "7", 324, 234, ArtifactoryRunningMode.HA, "1");
        ArtifactoryServer rt2 =
                new ArtifactoryServer(NODE_ID_2, 1612800485851L, "URL2", 2, ArtifactoryServerState.RUNNING,
                        ArtifactoryServerRole.MEMBER, 1612800485851L, "7", 324, 234, ArtifactoryRunningMode.HA, "2");
        List<ArtifactoryServer> activeMembers = Lists.newArrayList(rt1, rt2);
        when(artifactoryServersCommonService.getActiveMembers()).thenReturn(activeMembers);
    }

    @AfterMethod
    public void afterMethod() {
        ArtifactoryHome.unbind();
    }

    @Test
    public void testAssertIdIsNotEmpty_Empty() {
        assertThrows(BadRequestException.class, () -> liveLogsService.assertIdIsNotEmpty(""));
    }

    @Test
    public void testAssertIdIsNotEmpty_Null() {
        assertThrows(BadRequestException.class, () -> liveLogsService.assertIdIsNotEmpty(null));
    }

    @Test
    public void testAssertLiveLogsFeatureIsEnabled_EnterpriseLicense() {
        when(addonsManager.isEnterpriseLicensed()).thenReturn(true);
        liveLogsService.assertLiveLogsFeatureIsEnabled();
    }

    @Test
    public void testAssertLiveLogsFeatureIsEnabled_NoEnterpriseLicense() {
        when(addonsManager.isEnterpriseLicensed()).thenReturn(false);
        assertThrows(BadRequestException.class, () -> liveLogsService.assertLiveLogsFeatureIsEnabled());
    }

    @Test
    public void testGetLiveLogsConfig() {
        LiveLogsConfig liveLogsConfig = liveLogsService.getLiveLogsConfig();
        List<String> logs = liveLogsConfig.getLogs();
        List<String> nodes = liveLogsConfig.getNodes();
        long refreshRateMillis = liveLogsConfig.getRefreshRateMillis();
        assertEquals(logs.size(), 2);
        assertEquals(logs.get(0), ARTIFACTORY_REQUEST_LOG);
        assertEquals(logs.get(1), ARTIFACTORY_SERVICE_LOG);
        assertEquals(nodes.size(), 2);
        assertEquals(nodes.get(0), NODE_ID_1);
        assertEquals(nodes.get(1), NODE_ID_2);
        assertEquals(refreshRateMillis, 10000L);
    }

    @Test
    public void testGetLiveLogsData() {
        Mockito.doReturn(Collections.emptyList()).when(liveLogsService).getAdditionalSensitiveRegex();
        LiveLogsData liveLogsData = liveLogsService.getLiveLogsData(ARTIFACTORY_SERVICE_LOG, 0);
        assertNotNull(liveLogsData.getLastUpdateModified());
        assertNotNull(liveLogsData.getLastUpdateLabel());
        assertEquals(liveLogsData.getLogContent(), "");
        assertEquals(liveLogsData.getFileSize(), 0);
    }
}
