package org.artifactory.repo.db;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.xray.XrayAddon;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.StatusHolder;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.repo.InternalRepoPathFactory;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.snapshot.MavenSnapshotVersionAdapter;
import org.artifactory.request.Request;
import org.artifactory.request.RequestContext;
import org.artifactory.spring.InternalArtifactoryContext;
import org.testng.annotations.Test;

import java.util.Arrays;
import java.util.Enumeration;

import static org.easymock.EasyMock.anyObject;
import static org.easymock.EasyMock.createNiceMock;
import static org.easymock.EasyMock.eq;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.testng.Assert.assertEquals;

public class DbRepoDockerManifestDownloadTest {

    private AddonsManager addonsManager;

    @Test
    public void downloadUnscannedManifestAllowedOnlyForDockerRepoWhenHeaderSetToTrue() {
        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                true,
                false,
                "X-JFrog-Bypass-Unscanned-Blocking",
                "true",
                "manifest.json",
                repoType.equals(RepoType.Docker)));
    }

    @Test
    public void downloadWhenHeaderSetToFalse() {
        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                true,
                false,
                "X-JFrog-Bypass-Unscanned-Blocking",
                "false",
                "manifest.json",
                false));

        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                false,
                false,
                "X-JFrog-Bypass-Unscanned-Blocking",
                "false",
                "manifest.json",
                true));

    }

    @Test
    public void downloadUnscannedManifestWithWrongHeader() {
        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                true,
                false,
                "X-JFrog-Bypass-Download-Wrong",
                "false",
                "manifest.json",
                false));

        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                false,
                false,
                "X-JFrog-Bypass-Download-Wrong",
                "false",
                "manifest.json",
                true
        ));

    }

    @Test
    public void downloadUnscannedFileWhenHeaderSetToTrue() {
        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                true,
                false,
                "X-JFrog-Bypass-Unscanned-Blocking",
                "true",
                "other.file",
                false));


        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                false,
                false,
                "X-JFrog-Bypass-Unscanned-Blocking",
                "true",
                "other.file",
                true));

    }

    @Test
    public void downloadManifestWhenRequestIsNotSet() {
        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                true,
                true,
                "",
                "",
                "manifest.json",
                false));


        Arrays.stream(RepoType.values()).forEach(repoType -> verifyDownloadAllowed(
                repoType,
                false,
                true,
                "",
                "",
                "manifest.json",
                true));

    }

    public void verifyDownloadAllowed(
            RepoType repoType,
            boolean downloadBlockedByXRay,
            boolean requestIsNotSet,
            String headerName,
            String headerValue,
            String fileName,
            boolean downloadAllowed
    ) {
        setup();

        XrayAddon xrayAddon = createNiceMock(XrayAddon.class);
        expect(addonsManager.addonByType(XrayAddon.class)).andReturn(xrayAddon);
        expect(xrayAddon.isDownloadBlocked(anyObject(RepoPath.class), eq(true))).andReturn(false);
        expect(xrayAddon.isDownloadBlocked(anyObject(RepoPath.class), eq(false))).andReturn(downloadBlockedByXRay);
        replay(addonsManager, xrayAddon);
        RequestContext requestContext = createNiceMock(RequestContext.class);
        if (requestIsNotSet) {
            expect(requestContext.getRequest()).andReturn(null);
            replay(requestContext);
        } else {
            Request request = createNiceMock(Request.class);
            Enumeration<String> enumeration = createNiceMock(Enumeration.class);
            if (headerName.isEmpty()) {
                expect(enumeration.hasMoreElements()).andReturn(false);
            } else {
                expect(enumeration.hasMoreElements()).andReturn(true);
                expect(enumeration.nextElement()).andReturn(headerValue);
                expect(request.getHeaderValues(headerName)).andReturn(enumeration);
                expect(requestContext.getRequest()).andReturn(request);
            }
            replay(requestContext, enumeration, request);
        }
        RepoPath path = InternalRepoPathFactory.create("target", fileName);
        DbLocalRepo<LocalRepoDescriptor> repo = createDbRepo(repoType);
        StatusHolder status = repo.checkDownloadIsNotBlocked(requestContext, path);

        assertEquals(!status.isError(), downloadAllowed);

        cleanUp();
    }

    private void setup() {
        ArtifactoryContext context = createNiceMock(InternalArtifactoryContext.class);
        AuthorizationService authService = createNiceMock(AuthorizationService.class);
        addonsManager = createNiceMock(AddonsManager.class);
        expect(context.beanForType(AddonsManager.class)).andReturn(addonsManager);
        expect(context.getAuthorizationService()).andReturn(authService);
        ArtifactoryContextThreadBinder.bind(context);
        replay(context);
    }

    private void cleanUp() {
        ArtifactoryContextThreadBinder.unbind();
    }

    private DbLocalRepo<LocalRepoDescriptor> createDbRepo(RepoType repoType) {
        LocalRepoDescriptor descriptor = new LocalRepoDescriptor();
        descriptor.setType(repoType);
        return new DbLocalRepo<>(descriptor, null, null) {

            @Override
            public MavenSnapshotVersionAdapter getMavenSnapshotVersionAdapter() {
                return null;
            }

            @Override
            public boolean isSuppressPomConsistencyChecks() {
                return false;
            }
        };
    }

}
