/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo.service;

import org.apache.commons.compress.archivers.ArchiveInputStream;
import org.apache.commons.compress.archivers.zip.ZipArchiveInputStream;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.LocalCacheRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RealRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.mbean.MBeanRegistrationService;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.LocalCacheRepo;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RemoteRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.db.DbLocalRepo;
import org.artifactory.repo.http.mbean.HTTPConnectionPool;
import org.artifactory.repo.interceptor.StorageInterceptors;
import org.artifactory.repo.service.versioning.RepositoriesCache;
import org.artifactory.sapi.fs.VfsFile;
import org.artifactory.schedule.TaskService;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.storage.StorageService;
import org.artifactory.storage.db.security.service.VersioningCacheImpl;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.util.distribution.DistributionConstants;
import org.jetbrains.annotations.NotNull;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.springframework.util.ReflectionUtils;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import java.io.IOException;
import java.io.InputStream;
import java.lang.management.ManagementFactory;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import static org.junit.Assert.assertFalse;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

/**
 * @author Lior Gur
 */
public class RepositoryServiceImplTest extends ArtifactoryHomeBoundTest {

    private RepositoryServiceImpl repositoryService;

    @Mock
    private CentralConfigService centralConfigService;

    @Mock
    private TaskService taskService;

    @Mock
    private InternalArtifactoryContext artifactoryContext;

    @Mock
    private StorageInterceptors storageInterceptors;

    @Mock
    private VersioningCacheImpl<RepositoriesCache> cache;

    @Mock
    private ProjectsService projectsService;

    @Mock
    private StorageService storageService;

    @BeforeMethod
    public void init() {
        MockitoAnnotations.initMocks(this);
        repositoryService = Mockito.spy(new RepositoryServiceImpl());
    }

    @AfterMethod
    public void afterMethod() {
        ArtifactoryContextThreadBinder.unbind();
    }

    /**
     * RTFACT-14083
     * steps when deleting repository :
     * 1  blackout the repo in config descriptor
     * 2. save descriptor
     * 3. delete all files in militarisation
     * 4. remove repo from config descriptor
     * 5. save descriptor
     * 6. delete all files in repo (include repository himself)
     *
     * the test tests if step 4 is failing the repo remains blackedOut
     */
    @Test
    public void testRemoveRepository() {
        MutableCentralConfigDescriptor configDescriptor = getMutableCentralConfigDescriptor();
        configDescriptor.addLocalRepository(new LocalRepoDescriptor() {
            {
                setKey("testRepo");
                setType(RepoType.Generic);
            }
        });
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        when(artifactoryContext.beanForType(StorageInterceptors.class)).thenReturn(storageInterceptors);
        when(centralConfigService.getMutableDescriptor()).thenReturn(configDescriptor);

        assertFalse(configDescriptor.getLocalRepositoriesMap().get("testRepo").isBlackedOut());
        try {
            repositoryService.removeRepository("testRepo");
        } catch (NullPointerException ex) {
            //nothing to do here, we expected this exception
        }

        assertTrue(configDescriptor.getLocalRepositoriesMap().get("testRepo").isBlackedOut());
    }

    @NotNull
    private MutableCentralConfigDescriptor getMutableCentralConfigDescriptor() {
        Field field = ReflectionUtils.findField(RepositoryServiceImpl.class, "centralConfigService");
        ReflectionUtils.makeAccessible(field);
        ReflectionUtils.setField(field, repositoryService, centralConfigService);

        return new CentralConfigDescriptorImpl();
    }


    @Test(description = "Expects an exception while trying to get bean info after deleting orphan repositories",
            expectedExceptions = InstanceNotFoundException.class)
    public void testRemoveRemoteRepository()
            throws IntrospectionException, InstanceNotFoundException, ReflectionException {

        AolService aolService = mock(AolService.class);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);

        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        MutableCentralConfigDescriptor configDescriptor = getMutableCentralConfigDescriptor();
        MutableCentralConfigDescriptor oldConfigDescriptor = getMutableCentralConfigDescriptor();
        oldConfigDescriptor.addRemoteRepository(new HttpRepoDescriptor() {
            {
                setKey("testRepo");
                setType(RepoType.Generic);
            }
        });
        HashMap<String, LocalRepo> localRepoHashMap = new HashMap<>();
        HashMap<String, RemoteRepo> remoteRepoHashMap = new HashMap<>();
        remoteRepoHashMap.put("testRepo", mock(RemoteRepo.class));
        RepositoriesCache repositoriesCache =
                Mockito.spy(new RepositoriesCache(new HashMap<>(), new HashMap<>(), new HashMap<>(), new HashMap<>(),
                        remoteRepoHashMap, localRepoHashMap, null, null, null,
                        null, null));

        repositoryService.setTaskService(taskService);
        when(artifactoryContext.beanForType(AolService.class)).thenReturn(aolService);
        when(aolService.isAol()).thenReturn(false);
        doNothing().when(repositoryService).initClient();
        when(artifactoryContext.getCentralConfig()).thenReturn(centralConfigService);
        when(centralConfigService.getDescriptor()).thenReturn(configDescriptor);
        when(artifactoryContext.beanForType(InternalRepositoryService.class)).thenReturn(internalRepositoryService);
        when(internalRepositoryService.rebuildRepositoriesInTransaction()).thenReturn(repositoriesCache);

        repositoryService.init();
        ObjectName objName = MBeanRegistrationService.createObjectName("HTTPConnectionPool", "testRepo");
        HTTPConnectionPool pool = mock(HTTPConnectionPool.class);
        MBeanRegistrationService.register(pool, objName);
        assertTrue(ManagementFactory.getPlatformMBeanServer().isRegistered(objName));
        assertNotNull(ManagementFactory.getPlatformMBeanServer().getMBeanInfo(objName));
        repositoryService.deleteOrphanRepos(oldConfigDescriptor);

        //expects InstanceNotFoundException as mbean is no more registered
        ManagementFactory.getPlatformMBeanServer().getMBeanInfo(objName);
    }


    @Test
    public void testGotArchiveInputStream() throws IOException {
        RepoPath repoPath = RepoPathFactory.create("generic-local/bla.zip");
        InputStream inputStreamMock = prepMockForArchiveInputStreamTest(repoPath);

        ArchiveInputStream archiveInputStream = repositoryService.archiveInputStream(repoPath);

        assertTrue(archiveInputStream instanceof ZipArchiveInputStream);
        verify(inputStreamMock, times(0)).close();
    }

    @Test(expectedExceptions = IOException.class,
            expectedExceptionsMessageRegExp = "Could not process the archive format")
    public void testArchiveInputStreamClosed() throws IOException {
        RepoPath repoPath = RepoPathFactory.create("generic-local/bla.apk");
        InputStream inputStreamMock = prepMockForArchiveInputStreamTest(repoPath);
        try {
            repositoryService.archiveInputStream(repoPath);
        } finally {
            verify(inputStreamMock, times(1)).close();
        }
    }

    @Test
    public void testCreateProjectReposFilter() {
        initServices();
        String projectA = "proj1";
        List<String> allRepoKeys = List.of("repoA", "repoB", "repoC");
        when(projectsService.getProjectRepositoriesKeys(projectA, false)).thenReturn(Set.of("repoA", "repoB"));

        //Default project
        Function<String, Boolean> defaultFilter =
                repositoryService.createProjectReposFilter(ProjectsService.DEFAULT_PROJECT);
        List<String> defaultFilteredRepos =
                allRepoKeys.stream().filter(defaultFilter::apply).collect(Collectors.toList());
        assertEquals(defaultFilteredRepos.size(), 3);
        assertTrue(defaultFilteredRepos.contains("repoA"));
        assertTrue(defaultFilteredRepos.contains("repoB"));
        assertTrue(defaultFilteredRepos.contains("repoC"));

        //Project context
        Function<String, Boolean> projectFilter = repositoryService.createProjectReposFilter(projectA);
        List<String> projectFilteredRepos =
                allRepoKeys.stream().filter(projectFilter::apply).collect(Collectors.toList());
        assertEquals(projectFilteredRepos.size(), 2);
        assertTrue(projectFilteredRepos.contains("repoA"));
        assertTrue(projectFilteredRepos.contains("repoB"));
        assertFalse(projectFilteredRepos.contains("repoC"));
    }

    @Test
    public void testGetAllProjectSupportedRepoKeys() {
        initServices();
        HashMap<String, LocalRepo> localRepoHashMap = new HashMap<>();
        localRepoHashMap.put("1", createRepo("legit-repo1", RepoType.Maven));
        localRepoHashMap.put("2", createRepo("legit-repo2", RepoType.Generic));
        localRepoHashMap.put("3", createRepo("support", RepoType.Support));
        localRepoHashMap.put("4", createRepo("release-bundle", RepoType.ReleaseBundles));
        localRepoHashMap.put("5", createRepo("build-info", RepoType.BuildInfo));
        localRepoHashMap.put("6", createRepo("distribution", RepoType.Distribution));
        localRepoHashMap.put("7", createRepo(DistributionConstants.IN_TRANSIT_REPO_KEY, RepoType.Generic));
        RepositoriesCache repositoriesCache =
                Mockito.spy(new RepositoriesCache(new HashMap<>(), new HashMap<>(), new HashMap<>(), new HashMap<>(),
                        new HashMap<>(), localRepoHashMap, null, null, null,
                        null, null));
        when(cache.get()).thenReturn(repositoriesCache);

        // act
        Set<String> repoKeys = repositoryService.getAllProjectSupportedRepoKeys();
        assertEquals(repoKeys.size(), 2);
        assertTrue(repoKeys.contains("legit-repo1"));
        assertTrue(repoKeys.contains("legit-repo2"));
    }

    @Test
    public void testCachedReposNotFilteredOut() {
        final String projectKey = "testpr";
        final String repoKey = "test-repo-remote";
        final String repoKeyWithCacheSuffix = repoKey + "-cache";

        initServices();

        // 1. Mock repositoryService.getLocalAndCachedRepositories() results.
        LocalCacheRepo localCacheRepoMock = mock(LocalCacheRepo.class);
        when(localCacheRepoMock.getKey()).thenReturn(repoKeyWithCacheSuffix);
        LocalCacheRepoDescriptor repoDescriptor = new LocalCacheRepoDescriptor();
        repoDescriptor.setKey(repoKeyWithCacheSuffix);
        when(localCacheRepoMock.getDescriptor()).thenReturn(repoDescriptor);

        Map<String, LocalCacheRepo> localCacheRepositoriesMap = new HashMap<>();
        localCacheRepositoriesMap.put(repoKeyWithCacheSuffix, localCacheRepoMock);

        RepositoriesCache repositoriesCacheMock = new RepositoriesCache(new HashMap<>(), new HashMap<>(),
                new HashMap<>(), localCacheRepositoriesMap, new HashMap<>(), new HashMap<>(), mock(LocalRepo.class),
                new HashSet<>(), mock(LocalRepo.class), new HashMap<>(), new HashMap<>());
        when(cache.get()).thenReturn(repositoriesCacheMock);
        repositoryService.cache = this.cache;

        // 2. Mock repositoryService.createProjectReposFilter() results.
        when(projectsService.getProjectRepositoriesKeys(projectKey, false)).thenReturn(Set.of(repoKey));

        // 3. Assert RepositoryServiceImpl#getLocalAndCachedRepoDescriptorsByProject(projectKey) results.
        List<LocalRepoDescriptor> results = repositoryService.getLocalAndCachedRepoDescriptorsByProject(projectKey);
        // Make sure that the -cache repo is not filtered out.
        assertNotNull(results);
        assertEquals(results.size(), 1);
        assertEquals(results.get(0).getKey(), repoKeyWithCacheSuffix);

        // 4. Assert RepositoryServiceImpl#getCachedRepoDescriptorsByProject(projectKey) results.
        List<LocalCacheRepoDescriptor> results2 = repositoryService.getCachedRepoDescriptorsByProject(projectKey);
        // Make sure that the -cache repo is not filtered out.
        assertNotNull(results2);
        assertEquals(results2.size(), 1);
        assertEquals(results2.get(0).getKey(), repoKeyWithCacheSuffix);
    }

    @Test
    public void testIsInfoRepo() {
        assertTrue(repositoryService.isInfoRepo(RepoType.PipeInfo));
        assertTrue(repositoryService.isInfoRepo(RepoType.BuildInfo));
        assertFalse(repositoryService.isInfoRepo(RepoType.Generic));
        assertFalse(repositoryService.isInfoRepo(RepoType.Maven));
        assertFalse(repositoryService.isInfoRepo(RepoType.Docker));
        assertFalse(repositoryService.isInfoRepo(RepoType.ReleaseBundles));
    }

    @Test
    public void testGetLocalRepoDescriptors() {
        Map<String, LocalRepo> localRepositoriesMap = new HashMap<>();
        createLocalRepo("one", RepoType.PipeInfo, localRepositoriesMap);
        createLocalRepo("two", RepoType.Generic, localRepositoriesMap);
        createLocalRepo("three", RepoType.BuildInfo, localRepositoriesMap);

        RepositoriesCache repositoriesCacheMock = new RepositoriesCache(new HashMap<>(), new HashMap<>(),
                new HashMap<>(), new HashMap<>(), new HashMap<>(), localRepositoriesMap, mock(LocalRepo.class),
                new HashSet<>(), mock(LocalRepo.class), new HashMap<>(), new HashMap<>());

        when(cache.get()).thenReturn(repositoriesCacheMock);
        repositoryService.cache = this.cache;
        List<LocalRepoDescriptor> localRepoDescriptors = repositoryService.getLocalRepoDescriptors();
        Assert.assertEquals(localRepoDescriptors.size(), 1);
    }

    @Test
    public void testGetLocalRepoDescriptorsByProject() {
        initServices();
        Map<String, LocalRepo> localRepositoriesMap = new HashMap<>();
        createLocalRepo("one", RepoType.PipeInfo, localRepositoriesMap);
        createLocalRepo("two", RepoType.Generic, localRepositoriesMap);
        createLocalRepo("three", RepoType.BuildInfo, localRepositoriesMap);

        RepositoriesCache repositoriesCacheMock = new RepositoriesCache(new HashMap<>(), new HashMap<>(),
                new HashMap<>(), new HashMap<>(), new HashMap<>(), localRepositoriesMap, mock(LocalRepo.class),
                new HashSet<>(), mock(LocalRepo.class), new HashMap<>(), new HashMap<>());

        when(cache.get()).thenReturn(repositoriesCacheMock);
        repositoryService.cache = this.cache;
        String projectA = "projectA";
        when(projectsService.getProjectRepositoriesKeys(projectA, false)).thenReturn(Set.of("one", "two", "three"));
        List<LocalRepoDescriptor> localRepoDescriptors = repositoryService.getLocalRepoDescriptorsByProject(projectA);
        Assert.assertEquals(localRepoDescriptors.size(), 1);
    }

    private void createLocalRepo(String key, RepoType type, Map<String, LocalRepo> map) {
        LocalRepo<?> localRepo = mock(LocalRepo.class);
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setType(type);
        localRepoDescriptor.setKey(key);
        when((LocalRepoDescriptor) localRepo.getDescriptor()).thenReturn(localRepoDescriptor);
        map.put(key, localRepo);
    }

    private InputStream prepMockForArchiveInputStreamTest(RepoPath repoPath) {
        InputStream inputStreamMock = mock(InputStream.class);
        Map<String, LocalRepo> localRepositoriesMap = new HashMap<>();
        DbLocalRepo dbLocalRepoMock = mock(DbLocalRepo.class);
        VfsFile vfsFileMock = mock(VfsFile.class);
        when(vfsFileMock.getStream()).thenReturn(inputStreamMock);
        when(dbLocalRepoMock.getImmutableFile(repoPath)).thenReturn(vfsFileMock);
        localRepositoriesMap.put("generic-local", dbLocalRepoMock);
        RepositoriesCache repositoriesCacheMock = new RepositoriesCache(new HashMap<>(), new HashMap<>(),
                new HashMap<>(), new HashMap<>(), new HashMap<>(), localRepositoriesMap, mock(LocalRepo.class),
                new HashSet<>(), mock(LocalRepo.class), new HashMap<>(), new HashMap<>());
        when(cache.get()).thenReturn(repositoriesCacheMock);
        repositoryService.cache = this.cache;
        return inputStreamMock;
    }

    private void initServices() {
        Field field = ReflectionUtils.findField(RepositoryServiceImpl.class, "projectsService");
        ReflectionUtils.makeAccessible(field);
        ReflectionUtils.setField(field, repositoryService, projectsService);
        field = ReflectionUtils.findField(RepositoryServiceImpl.class, "cache");
        ReflectionUtils.makeAccessible(field);
        ReflectionUtils.setField(field, repositoryService, cache);
    }

    private LocalRepo createRepo(String repoKey, RepoType repoType) {
        LocalRepo localRepo = mock(LocalRepo.class);
        when(localRepo.getKey()).thenReturn(repoKey);
        RealRepoDescriptor repoDescriptor = spy(new LocalRepoDescriptor());
        when(repoDescriptor.getKey()).thenReturn(repoKey);
        when(repoDescriptor.getType()).thenReturn(repoType);
        when(localRepo.getDescriptor()).thenReturn(repoDescriptor);
        return localRepo;
    }
}