/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.security.access;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.security.SecurityDescriptor;
import org.artifactory.security.UserInfo;
import org.jfrog.access.client.AccessClient;
import org.jfrog.access.client.AccessClientHttpException;
import org.jfrog.access.client.auth.AuthClient;
import org.jfrog.access.model.MessageModelCode;
import org.jfrog.access.rest.user.LoginRequest;
import org.jfrog.access.rest.user.UserWithGroupsResponse;
import org.jfrog.access.user.UserMFAStatus;
import org.springframework.security.authentication.AuthenticationServiceException;
import org.springframework.security.authentication.CredentialsExpiredException;
import org.springframework.security.authentication.LockedException;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.testng.annotations.AfterClass;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.HashMap;

import static org.easymock.EasyMock.anyObject;
import static org.easymock.EasyMock.createMock;
import static org.easymock.EasyMock.createStrictMock;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.easymock.EasyMock.reset;
import static org.easymock.EasyMock.verify;
import static org.testng.Assert.assertTrue;

/**
 * @author Noam Shemesh
 */
public class AccessUserPassAuthenticationProviderTest {

    private AccessUserPassAuthenticationProvider accessUserPassAuthenticationProvider;
    private AccessClient accessClient;
    private AuthClient authClient;
    private AccessService accessService;
    private CentralConfigService centralConfigService;
    private SecurityDescriptor securityDescriptor;
    private UsernamePasswordAuthenticationToken defaultAuth;

    @BeforeClass
    public void before() {
        authClient = createMock(AuthClient.class);
        accessClient = createMock(AccessClient.class);
        accessService = createStrictMock(AccessService.class);
        centralConfigService = createMock(CentralConfigService.class);
        final CentralConfigDescriptor configDescriptor = createMock(CentralConfigDescriptor.class);
        securityDescriptor = createMock(SecurityDescriptor.class);
        expect(configDescriptor.getSecurity()).andReturn(securityDescriptor).anyTimes();
        expect(centralConfigService.getDescriptor()).andReturn(configDescriptor).anyTimes();

        expect(accessClient.auth()).andReturn(authClient).anyTimes();
        expect(accessService.getAccessClient()).andReturn(accessClient).anyTimes();

        replay(accessClient, accessService, centralConfigService, configDescriptor);

        accessUserPassAuthenticationProvider = new AccessUserPassAuthenticationProvider(accessService, null,
                centralConfigService);
    }

    @AfterClass
    public void after() {
        verify(accessClient, accessService);
    }

    @BeforeMethod
    public void beforeMethod() {
        defaultAuth = new UsernamePasswordAuthenticationToken("noam", "psswrd");
    }

    @AfterMethod
    public void afterMethod() {
        verify(authClient);
        reset(authClient, securityDescriptor);
    }

    @Test
    public void testAuthenticateSuccess() {
        expect(authClient.authenticate(anyObject(LoginRequest.class))).andReturn(
                new UserWithGroupsResponse("noam")
                        .customData(new HashMap<>())
                        .mfaStatus(UserMFAStatus.NONE)
        ).once();
        replay(authClient);

        Authentication auth = accessUserPassAuthenticationProvider.authenticate(defaultAuth);

        assertTrue(auth.isAuthenticated());
    }

    @Test(expectedExceptions = AuthenticationServiceException.class)
    public void testAuthenticateFailed() {
        String errorMessage = "not authed";
        expect(authClient.authenticate(anyObject(LoginRequest.class)))
                .andThrow(new AccessClientHttpException(401, errorMessage)).once();
        expect(securityDescriptor.isAnyExternalAuthEnabled()).andReturn(false);
        replay(authClient, securityDescriptor);

        accessUserPassAuthenticationProvider.authenticate(defaultAuth);
    }

    @Test(expectedExceptions = LockedException.class)
    public void testAuthenticateForbidden() {
        String errorMessage = "{ \"errors\": [ { \"code\": \"" + MessageModelCode.CODE_TOO_MANY_REQUESTS + "\" } ] }";
        expect(authClient.authenticate(anyObject(LoginRequest.class)))
                .andThrow(new AccessClientHttpException(429, errorMessage)).once();
        expect(securityDescriptor.isAnyExternalAuthEnabled()).andReturn(false);
        replay(authClient, securityDescriptor);

        accessUserPassAuthenticationProvider.authenticate(defaultAuth);
    }

    @Test(expectedExceptions = CredentialsExpiredException.class)
    public void testCredentialsExpired() {
        String errorMessage = "{ \"errors\": [ { \"code\": \"" + MessageModelCode.CODE_PASSWORD_EXPIRED + "\" } ] }";
        expect(authClient.authenticate(anyObject(LoginRequest.class))).andThrow(new AccessClientHttpException(403,
                errorMessage)).once();
        replay(authClient);

        accessUserPassAuthenticationProvider.authenticate(defaultAuth);
    }

    @Test(expectedExceptions = AuthenticationServiceException.class)
    public void shouldNotAuthenticateForEnabledExternalAuthAndDisabledBasicAuth() {
        expect(securityDescriptor.isAnyExternalAuthEnabled()).andReturn(true);
        expect(securityDescriptor.isBasicAuthEnabled()).andReturn(false);
        replay(authClient, securityDescriptor);

        accessUserPassAuthenticationProvider.authenticate(defaultAuth);
    }

    @Test
    public void shouldAuthenticateAnonymousUserForEnabledExternalAuthAndDisabledBasicAuth() {
        expect(authClient.authenticate(anyObject(LoginRequest.class))).andReturn(
                new UserWithGroupsResponse(UserInfo.ANONYMOUS)
                        .customData(new HashMap<>())
                        .mfaStatus(UserMFAStatus.NONE)
        ).once();

        expect(securityDescriptor.isAnyExternalAuthEnabled()).andReturn(true);
        expect(securityDescriptor.isBasicAuthEnabled()).andReturn(false);
        replay(authClient, securityDescriptor);

        final Authentication authenticate = accessUserPassAuthenticationProvider.authenticate(
                new UsernamePasswordAuthenticationToken(UserInfo.ANONYMOUS, ""));
        assertTrue(authenticate.isAuthenticated());
    }

    @Test
    public void shouldAuthenticateForEnabledExternalAuthAndEnabledBasicAuth() {
        expect(authClient.authenticate(anyObject(LoginRequest.class))).andReturn(
                new UserWithGroupsResponse("noam")
                        .customData(new HashMap<>())
                        .mfaStatus(UserMFAStatus.NONE)
        ).once();
        expect(securityDescriptor.isAnyExternalAuthEnabled()).andReturn(true);
        expect(securityDescriptor.isBasicAuthEnabled()).andReturn(true);
        replay(authClient, securityDescriptor);

        final Authentication authenticate = accessUserPassAuthenticationProvider.authenticate(defaultAuth);
        assertTrue(authenticate.isAuthenticated());
    }

    @Test
    public void shouldAuthenticateForDisabledExternalAuthAndEnabledBasicAuth() {
        expect(authClient.authenticate(anyObject(LoginRequest.class))).andReturn(
                new UserWithGroupsResponse("noam")
                        .customData(new HashMap<>())
                        .mfaStatus(UserMFAStatus.NONE)
        ).once();
        expect(securityDescriptor.isAnyExternalAuthEnabled()).andReturn(false);
        expect(securityDescriptor.isBasicAuthEnabled()).andReturn(true);
        replay(authClient, securityDescriptor);

        final Authentication authenticate = accessUserPassAuthenticationProvider.authenticate(defaultAuth);
        assertTrue(authenticate.isAuthenticated());
    }
}