package org.artifactory.storage.db.environment.converter.shared.version.v3;

import lombok.SneakyThrows;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.environment.converter.BasicEnvironmentConverter;
import org.artifactory.storage.event.EventType;
import org.artifactory.storage.replication.errors.InternalReplicationEventType;
import org.artifactory.version.ArtifactoryVersionProvider;
import org.artifactory.version.CompoundVersionDetails;
import org.jfrog.config.db.TemporaryDBChannel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.jdbc.core.JdbcTemplate;

import java.util.Map;

/**
 * @author Mor Merhav
 */
public class TaskTypeToNodeEventTypeConverter implements BasicEnvironmentConverter {
    private static final Logger log = LoggerFactory.getLogger(TaskTypeToNodeEventTypeConverter.class);

    private static final Map<Integer, Integer> taskTypeCodeToEventTypeCodeMap =
            Map.of(InternalReplicationEventType.DEPLOY.code(), EventType.create.code(),
                    InternalReplicationEventType.MKDIR.code(), EventType.create.code(),
                    InternalReplicationEventType.DELETE.code(), EventType.delete.code(),
                    InternalReplicationEventType.PROPERTY_CHANGE.code(), EventType.props.code());

    @Override
    public boolean isInterested(ArtifactoryHome home, CompoundVersionDetails source, CompoundVersionDetails target) {
        return source.getVersion().before(ArtifactoryVersionProvider.v770.get());
    }

    @SneakyThrows
    @Override
    public void convert(ArtifactoryHome artifactoryHome, CompoundVersionDetails source, CompoundVersionDetails target) {
        try (TemporaryDBChannel temporaryDBChannel = new TemporaryDBChannel(artifactoryHome.getDBProperties())) {
            JdbcTemplate jdbcTemplate = temporaryDBChannel.asJdbcTemplate();
            log.info("Starting to convert replication-specific task_type to event_type");
            taskTypeCodeToEventTypeCodeMap.entrySet().forEach(entry -> toEventType(jdbcTemplate, entry));
            log.info("Finished converting replication-specific task_type to event_type");
        }
    }

    @Override
    public void assertConversionPreconditions(ArtifactoryHome home) {
        // noop
    }

    private void toEventType(JdbcTemplate jdbcTemplate, Map.Entry<Integer, Integer> taskTypeToEventTypeEntry) {
        log.info("Converting {} to {}", taskTypeToEventTypeEntry.getKey(), taskTypeToEventTypeEntry.getValue());
        jdbcTemplate.update("UPDATE node_events_errors SET event_type = ? WHERE event_type = ?",
                taskTypeToEventTypeEntry.getValue(),
                taskTypeToEventTypeEntry.getKey());
        log.info("Converted {} to {}", taskTypeToEventTypeEntry.getKey(), taskTypeToEventTypeEntry.getValue());
    }
}
