/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.security.service.access;

import com.google.common.collect.ImmutableList;
import org.artifactory.api.security.AclService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.model.xstream.security.RepoPermissionTargetImpl;
import org.artifactory.security.AceInfo;
import org.artifactory.security.Acl;
import org.artifactory.security.MutableAceInfo;
import org.artifactory.security.PermissionTarget;
import org.artifactory.security.RepoAcl;
import org.artifactory.security.RepoPermissionTarget;
import org.artifactory.security.access.AccessService;
import org.artifactory.test.ArtifactoryHomeStub;
import org.jetbrains.annotations.Nullable;
import org.jfrog.access.client.AccessClient;
import org.jfrog.access.client.permission.PermissionsClient;
import org.jfrog.access.common.ResourceType;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.rest.permission.PermissionActionsResponse;
import org.jfrog.access.rest.permission.PermissionResponse;
import org.jfrog.access.rest.permission.PermissionsResponse;
import org.jfrog.common.ClockUtils;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.Collections;
import java.util.Set;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.clearInvocations;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;

/**
 * @author Noam Shemesh
 */
public class AccessAclStoreServiceTest {
    private AclStoreServiceImpl aclStoreServiceImpl;

    @Mock
    private AccessService accessService;
    @Mock
    private PermissionsClient permissionsClient;
    @Mock
    private AclService aclService;

    @BeforeMethod
    public void beforeTest() {
        MockitoAnnotations.initMocks(this);

        ArtifactoryHome.bind(new ArtifactoryHomeStub());
        aclStoreServiceImpl = new AclStoreServiceImpl();
        aclStoreServiceImpl.setAccessService(accessService);
        aclStoreServiceImpl.setAclService(aclService);
        aclStoreServiceImpl.init();

        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.aclMinimumWaitRefreshDataSeconds.getPropertyName(), "6");
        long time = ClockUtils.epochMillis();
        AccessClient accessClient = mock(AccessClient.class);
        when(accessClient.permissions()).thenReturn(permissionsClient);
        PermissionResponse permissionResponse = new PermissionResponse()
                .modified(time - ConstantValues.aclMinimumWaitRefreshDataSeconds.getLong() - 1)
                .actions(new PermissionActionsResponse())
                .resourceType(ResourceType.REPO).serviceId("ser").id("name");
        when(permissionsClient.findPermissionsByResourceType(any()))
                .thenReturn(new PermissionsResponse().permissions(ImmutableList.of(permissionResponse)));
        when(permissionsClient.lastUpdated()).thenReturn(time);

        when(accessService.getAccessClient()).thenReturn(accessClient);
        when(accessService.getArtifactoryServiceId()).thenReturn(ServiceId.fromFormattedName("jfrt@service"));
    }

    @Test(enabled = false, description = "flaky")
    public void testRefreshingCache() throws InterruptedException {
        aclStoreServiceImpl.getAclCache();
        aclStoreServiceImpl.getAclCache();
        Thread.sleep(6500);
        aclStoreServiceImpl.getAclCache();

        verify(permissionsClient, times(2)).lastUpdated();
    }

    @Test(enabled = false, description = "flaky")
    public void testRefreshingCacheForAclMap() throws InterruptedException {
        aclStoreServiceImpl.getAllRepoAcls();
        aclStoreServiceImpl.getAllRepoAcls();
        Thread.sleep(6500);
        aclStoreServiceImpl.getAllRepoAcls();

        verify(permissionsClient, times(2)).lastUpdated();
    }

    @Test
    public void testUpdateAclIsSkippedForInternal() {
        Acl<? extends PermissionTarget> acl = new MockAcl();
        when(aclService.isInternalAcl(acl)).thenReturn(false);
        aclStoreServiceImpl.updateAcl(acl);
        verify(permissionsClient).replacePermission(any());
        clearInvocations(permissionsClient);

        when(aclService.isInternalAcl(acl)).thenReturn(true);
        aclStoreServiceImpl.updateAcl(acl);
        verifyNoMoreInteractions(permissionsClient);
    }

    public static class MockAcl implements RepoAcl {

        @Override
        public RepoPermissionTarget getPermissionTarget() {
            return new RepoPermissionTargetImpl("permissionTarget");
        }

        @Override
        public Set<AceInfo> getAces() {
            return Collections.emptySet();
        }

        @Override
        public Set<MutableAceInfo> getMutableAces() {
            return Collections.emptySet();
        }

        @Override
        public String getUpdatedBy() {
            return null;
        }

        @Override
        public long getLastUpdated() {
            return 0;
        }

        @Nullable
        @Override
        public String getAccessIdentifier() {
            return "identifier";
        }
    }
}