/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.admin.configuration.repositories.util;

import lombok.AllArgsConstructor;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.projects.ProjectsService;
import org.artifactory.projects.ProjectsUtils;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.rest.model.admin.configuration.repository.info.RepositoryInfo;
import org.artifactory.ui.rest.model.admin.configuration.repository.info.RepositoryInfoListFactory;
import org.artifactory.ui.utils.RequestUtils;
import org.jfrog.access.client.project.ProjectResourceEntity;
import org.jfrog.access.client.project.ProjectResourceStatus;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.Nullable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * @author Aviad Shikloshi
 */
@Slf4j
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
@AllArgsConstructor
@NoArgsConstructor
public class GetRepositoryInfoService implements RestService {

    @Autowired
    private RepositoryService repositoryService;

    @Autowired
    private CentralConfigService configService;

    @Autowired
    private ProjectsService projectsService;

    @Autowired
    private AuthorizationService authorizationService;

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        String repoType = request.getPathParamByKey("repoType");
        String projectKey = RequestUtils.getProjectKeyFromRequest(request);
        Function<RepositoryInfo, Boolean> filterProjectAdminRepositories =
                createFilterProjectAdminRepositoriesFunction(projectKey);
        List<RepositoryInfo> repositoryInfoListInfo =
                RepositoryInfoListFactory
                        .createRepositoryInfo(repoType, configService, repositoryService).stream()
                        .filter(filterProjectAdminRepositories::apply)
                        .collect(Collectors.toList());
        addProjectRelatedData(projectKey, repositoryInfoListInfo);
        response.iModelList(repositoryInfoListInfo);
    }

    private void addProjectRelatedData(String projectKey, List<RepositoryInfo> repositoryInfoListInfo) {
        Set<String> repoKeys =
                repositoryInfoListInfo.stream().map(RepositoryInfo::getRepoKey).collect(Collectors.toSet());
        Map<String, String> repoKeyToProjectMap = projectsService.getRepoKeyToProjectMap(repoKeys, true);
        if (ProjectsUtils.isDefaultProject(projectKey) || ProjectsUtils.isUnassignedContext(projectKey)) {
            Map<String, Map<String, ProjectResourceEntity>> projectsResources =
                    projectsService.getAllProjectsRepositoriesMap(true);
            repositoryInfoListInfo.forEach(repositoryInfo -> {
                String repoProjectKey =
                        repoKeyToProjectMap.getOrDefault(repositoryInfo.getRepoKey(), ProjectsService.DEFAULT_PROJECT);
                ProjectResourceEntity resourceEntity =
                        projectsResources.getOrDefault(repoProjectKey, Map.of()).get(repositoryInfo.getRepoKey());
                setProjectsData(repoProjectKey, repositoryInfo, resourceEntity);
            });
        } else {
            Map<String, ProjectResourceEntity> projectResources =
                    projectsService.getProjectRepositoriesMap(projectKey, true);
            repositoryInfoListInfo.forEach(repositoryInfo -> {
                String repoProjectKey =
                        repoKeyToProjectMap.getOrDefault(repositoryInfo.getRepoKey(), ProjectsService.DEFAULT_PROJECT);
                ProjectResourceEntity resourceEntity = projectResources.get(repositoryInfo.getRepoKey());
                setProjectsData(repoProjectKey, repositoryInfo, resourceEntity);
            });
        }
    }

    private void setProjectsData(@NonNull String projectKey, @NonNull RepositoryInfo repositoryInfo,
            @Nullable ProjectResourceEntity resourceEntity) {
        repositoryInfo.setProjectKey(projectKey);
        repositoryInfo.setProjectName(projectsService.getProjectDisplayName(projectKey));
        log.debug("Setting projectKey {} and projectName {} to repo {}", projectKey, repositoryInfo.getProjectName(),
                repositoryInfo.getRepoKey());
        if (resourceEntity == null) {
            return;
        }
        if (resourceEntity.getResourceStatus().equals(ProjectResourceStatus.SHARED_WITH_ALL)) {
            repositoryInfo.setShareWithAllProjects(true);
        }
        repositoryInfo.setSharedWithProjects(resourceEntity.getSharedWithProjects());
        log.debug("Setting shared data for repo {}. ShareWithAll: {}, SharedWithProjects: {}",
                repositoryInfo.getRepoKey(), repositoryInfo.isShareWithAllProjects(),
                repositoryInfo.getSharedWithProjects());
    }

    @NonNull
    private Function<RepositoryInfo, Boolean> createFilterProjectAdminRepositoriesFunction(
            String projectKey) {
        Function<RepositoryInfo, Boolean> containsFunction;
        if (ProjectsUtils.isDefaultProject(projectKey)) {
            boolean isAdmin = authorizationService.isAdmin();
            containsFunction = (repositoryInfo) -> isAdmin;
        } else if (ProjectsUtils.isUnassignedContext(projectKey)) {
            boolean isAdmin = authorizationService.isAdmin();
            Set<String> repositories = getUnassignedRepositories();
            log.debug("Limiting the repoList to contain only unassigned repositories within the list: {}",
                    repositories);
            containsFunction = repositoryInfo -> isAdmin && repositories.contains(repositoryInfo.getRepoKey());
        } else if (authorizationService.isProjectAdmin(projectKey)) {
            Set<String> repositories = projectsService.getProjectRepositoriesKeys(projectKey, true);
            log.debug("Limiting the repoList to contain only repositories from project {} within the list: {}",
                    projectKey, repositories);
            containsFunction = repositoryInfo -> repositories.contains(repositoryInfo.getRepoKey());
        } else {
            containsFunction = (repositoryInfo) -> false;
        }
        return containsFunction;
    }

    private Set<String> getUnassignedRepositories() {
        Map<String, String> repoToProjectMap =
                projectsService.getRepoKeyToProjectMap(repositoryService.getAllRepoKeys(), true);
        return repoToProjectMap.entrySet()
                .stream()
                .filter(entry -> ProjectsUtils.isDefaultProject(entry.getValue()))
                .map(Map.Entry::getKey)
                .collect(Collectors.toSet());
    }
}
