/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2020 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.admin.security.keys.keypairs;

import org.apache.commons.lang.StringUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.rest.keys.KeyPairModel;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.KeyPairDescriptor;
import org.artifactory.keys.KeyPairsUtil;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.rest.exception.BadRequestException;
import org.jfrog.security.crypto.signing.rsa.RSAKeyParser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import static org.artifactory.descriptor.repo.KeyPairDescriptor.RSA_TYPE;

/**
 * @author Yuval Reches
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class CreateKeyPairService implements RestService {
    private static final Logger log = LoggerFactory.getLogger(CreateKeyPairService.class);

    private final CentralConfigService centralConfigService;

    @Autowired
    public CreateKeyPairService(CentralConfigService centralConfigService) {
        this.centralConfigService = centralConfigService;
    }

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        try {
            KeyPairModel requestModel = (KeyPairModel) request.getImodel();
            requestModel.validate();
            validatePerType(requestModel);
            log.debug("Create Key Pair request for {}", requestModel.getPairName());
            KeyPairDescriptor keyPairDescriptor = KeyPairsUtil.modelToDescriptor(requestModel);
            MutableCentralConfigDescriptor mutableDescriptor = centralConfigService.getMutableDescriptor();
            mutableDescriptor.addKeyPair(keyPairDescriptor);
            centralConfigService.saveEditedDescriptorAndReload(mutableDescriptor, keyPairDescriptor);
            response.info("Successfully added Key Pair " + requestModel.getPairName());
        } catch (Exception e) {
            response.error("Failed to create a Key Pair: " + e.getMessage());
        }
    }

    /**
     * @throws Exception in case the request private/public keys are not in a valid format.
     */
    private void validatePerType(KeyPairModel request) throws Exception {
        if (StringUtils.equalsIgnoreCase(RSA_TYPE, request.getPairType())) {
            RSAKeyParser.verifyRSAKeyPair(request.getPrivateKey(), request.getPublicKey(), request.getPassphrase());
        } else {
            throw new BadRequestException(
                    "Key Pair supported type is RSA only, request type is: " + request.getPairType());
        }
    }
}