package org.artifactory.ui.rest.service.artifacts.browse.treebrowser.tabs.permission;

import org.artifactory.aol.AolService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.tabs.permission.RepoPermissionsModel;
import org.artifactory.ui.rest.service.builds.buildsinfo.tabs.permissions.GetBuildEffectivePermissionService;
import org.artifactory.ui.rest.service.distribution.GetReleaseBundleEffectivePermissionService;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.HashMap;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;

/**
 * @author Yevdo Abramov
 * Created on 27/10/2020
 */
@Test
public class GetRepoEffectivePermissionServiceTest {
    @Mock
    private AuthorizationService authService;
    @Mock
    private RepositoryService repoService;
    @Mock
    private GetBuildEffectivePermissionService buildEffectivePermissionService;
    @Mock
    private GetReleaseBundleEffectivePermissionService releaseBundleEffectivePermissionService;
    @Mock
    private UserGroupService userGroupService;

    private GetRepoEffectivePermissionService service;

    @BeforeMethod
    public void setUp() {
        MockitoAnnotations.openMocks(this);
        service = new GetRepoEffectivePermissionService(authService, repoService, userGroupService,
                buildEffectivePermissionService, releaseBundleEffectivePermissionService);
    }

    @Test
    public void testExecuteWhenNoPermissions() {
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        RestResponse response = mock(RestResponse.class);

        when(request.getQueryParamByKey("repoKey")).thenReturn("my_repo");
        when(request.getQueryParamByKey("path")).thenReturn("my_path");
        when(authService.canManage(any(RepoPath.class))).thenReturn(false);

        service.execute(request, response);

        verify(request).getQueryParamByKey("repoKey");
        verify(request).getQueryParamByKey("path");
        verify(authService).canManage(any());
        verifyNoMoreInteractions(userGroupService);
    }

    @Test
    public void testExecuteIsAcceptingManagePermissions() {
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        RestResponse response = mock(RestResponse.class);

        when(request.getQueryParamByKey("repoKey")).thenReturn("my_repo");
        when(request.getQueryParamByKey("path")).thenReturn("my_path");
        when(authService.canManage(any(RepoPath.class))).thenReturn(true);

        service.execute(request, response);

        verify(request).getQueryParamByKey("repoKey");
        verify(request).getQueryParamByKey("path");
        verify(authService).canManage(any());
        verify(userGroupService).getRepoPathAcls(any());
    }

    @Test
    public void testExecutePipelineInfoRepo() {
        ArtifactoryRestRequest<?> request = mock(ArtifactoryRestRequest.class);
        RestResponse<?> response = mock(RestResponse.class);
        doAnswer(invocationOnMock -> {
            Assert.assertEquals(invocationOnMock.getArguments().length, 1);
            RepoPermissionsModel repoPermissionsModel = invocationOnMock.getArgument(0, RepoPermissionsModel.class);
            Assert.assertNotNull(repoPermissionsModel);
            Assert.assertEquals(repoPermissionsModel.getPermissionTargets().size(), 0);
            Assert.assertEquals(repoPermissionsModel.getGroupEffectivePermissions().size(), 1);
            Assert.assertEquals(repoPermissionsModel.getUserEffectivePermissions().size(), 1);
            return null;
        }).when(response).iModel(any());

        when(request.getQueryParamByKey("repoKey")).thenReturn("my_repo");
        when(request.getQueryParamByKey("path")).thenReturn("my_path");
        // Repo type of pipe info
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setType(RepoType.PipeInfo);
        when(repoService.localRepoDescriptorByKey("my_repo")).thenReturn(localRepoDescriptor);

        // Security mocking
        when(authService.canManage(any(RepoPath.class))).thenReturn(true);
        HashMap<String, Boolean> adminUserMap = new HashMap<>();
        adminUserMap.put("test", true);
        when(userGroupService.getAllUsersAndAdminStatus(true)).thenReturn(adminUserMap);
        ArrayList<String> adminGroups = new ArrayList<>();
        adminGroups.add("groupAdmin");
        when(userGroupService.getAllAdminGroupsNames()).thenReturn(adminGroups);

        // Context creation
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        AolService aolService = mock(AolService.class);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        ArtifactoryContextThreadBinder.bind(context);

        service.execute(request, response);

        verify(request).getQueryParamByKey("repoKey");
        verify(request).getQueryParamByKey("path");
        verify(authService).canManage(any());
    }

    @AfterMethod
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void testBuildEffectivePermissionsExcludingInternals() {
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        RestResponse response = mock(RestResponse.class);
        String buildRepoKey = "proj-build-info";
        RepoPath repoPath = RepoPathFactory.create(buildRepoKey, "my_path");
        when(request.getQueryParamByKey("repoKey")).thenReturn(buildRepoKey);
        when(request.getQueryParamByKey("path")).thenReturn(repoPath.getPath());
        when(authService.canManage(any(RepoPath.class))).thenReturn(true);
        LocalRepoDescriptor buildRepo = new LocalRepoDescriptor();
        buildRepo.setKey(buildRepoKey);
        buildRepo.setType(RepoType.BuildInfo);
        when(repoService.localRepoDescriptorByKey(buildRepoKey)).thenReturn(buildRepo);
        service.execute(request, response);

        verify(request).getQueryParamByKey("repoKey");
        verify(request).getQueryParamByKey("path");
        verify(authService).canManage(any());
        verify(buildEffectivePermissionService).getPermissionModel(repoPath);
    }
}