package org.artifactory.ui.rest.service.utils.repositories;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.projects.ProjectsService;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.ui.rest.model.utils.repositories.RepositoriesData;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.mockito.ArgumentMatchers;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.collections.Lists;

import java.util.Arrays;
import java.util.Collection;
import java.util.Map;
import java.util.Set;

@RunWith(Parameterized.class)
public class GetAllRepositoriesServiceTest {

    private static final String PROJECT_A = "projectA";
    private static final String LOCAL_A = "local-repo";
    private static final String LOCAL_PROJ = "prj-local-repo";

    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private RepositoryService repoService;
    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private ProjectsService projectsService;
    @Mock
    private ArtifactoryRestRequest request;
    private ArtifactoryRestResponse response;
    private GetAllRepositoriesService getAllRepositoriesService;

    private final String action;
    private final String listTypeToTest;
    private final int repoCountOnDefault;

    public GetAllRepositoriesServiceTest(String action, String listTypeToTest, int repoCountOnDefault) {
        this.action = action;
        this.listTypeToTest = listTypeToTest;
        this.repoCountOnDefault = repoCountOnDefault;
    }

    @Before
    public void setUp() {
        MockitoAnnotations.openMocks(this);
        response = new ArtifactoryRestResponse();
        getAllRepositoriesService = Mockito.spy(new GetAllRepositoriesService());
        getAllRepositoriesService.setAuthorizationService(authorizationService);
        getAllRepositoriesService.setRepoService(repoService);
        getAllRepositoriesService.setCentralConfigService(centralConfigService);
        getAllRepositoriesService.setProjectsService(projectsService);

        final Set<String> projectRepos = Set.of(LOCAL_PROJ);
        Mockito.when(authorizationService.canRead(ArgumentMatchers.any())).thenReturn(true);
        Mockito.when(projectsService.getProjectRepositoriesKeys(PROJECT_A, false)).thenReturn(projectRepos);
        Mockito.when(repoService.getLocalAndCachedRepoDescriptorsByProject(ProjectsService.DEFAULT_PROJECT))
                .thenReturn(Lists.newArrayList(createLocalRepoDescriptor(LOCAL_A),
                        createLocalRepoDescriptor(LOCAL_PROJ)));
        Mockito.when(repoService.getLocalAndCachedRepoDescriptorsByProject(PROJECT_A))
                .thenReturn(Lists.newArrayList(createLocalRepoDescriptor(LOCAL_PROJ)));
        Mockito.when(repoService.getLocalAndRemoteRepoDescriptorsByProject(PROJECT_A))
                .thenReturn(Lists.newArrayList(createLocalRepoDescriptor(LOCAL_PROJ),
                        createLocalRepoDescriptor("pipe-info", RepoType.PipeInfo)));
        Mockito.when(repoService.getLocalAndRemoteRepoDescriptorsByProject(ProjectsService.DEFAULT_PROJECT))
                .thenReturn(Lists.newArrayList(createLocalRepoDescriptor(LOCAL_A),
                        createLocalRepoDescriptor(LOCAL_PROJ)));
        Mockito.when(repoService.isInfoRepo(RepoType.PipeInfo)).thenReturn(true);
        Mockito.when(authorizationService.isAdmin()).thenReturn(true);
        MutableCentralConfigDescriptor configDescriptor = new CentralConfigDescriptorImpl();
        configDescriptor.setFileUploadMaxSizeMb(1);
        configDescriptor.setLocalRepositoriesMap(Map.of(LOCAL_A, createLocalRepoDescriptor(LOCAL_A), LOCAL_PROJ,
                createLocalRepoDescriptor(LOCAL_PROJ)));
        Mockito.when(centralConfigService.getMutableDescriptor()).thenReturn(configDescriptor);
        Mockito.when(centralConfigService.getDescriptor()).thenReturn(configDescriptor);
    }

    @Test
    public void testExecuteWithDifferentActions() {
        Mockito.when(request.getQueryParamByKey(action)).thenReturn("true");

        // Default
        Mockito.when(request.getQueryParamByKey(ProjectsService.PROJECT_KEY)).thenReturn("");
        getAllRepositoriesService.execute(request, response);
        RepositoriesData repoData = (RepositoriesData) response.getIModel();
        if ("repoTypesList".equals(listTypeToTest)) {
            Assert.assertEquals(repoCountOnDefault, repoData.getRepoTypesList().size());
            Assert.assertNull(repoData.getRepoList());
        } else {
            Assert.assertEquals(repoCountOnDefault, repoData.getRepoList().size());
            Assert.assertNull(repoData.getRepoTypesList());
        }

        // Project
        Mockito.when(request.getQueryParamByKey(ProjectsService.PROJECT_KEY)).thenReturn(PROJECT_A);
        response = new ArtifactoryRestResponse();
        getAllRepositoriesService.execute(request, response);
        repoData = (RepositoriesData) response.getIModel();
        if ("repoTypesList".equals(listTypeToTest)) {
            Assert.assertEquals(1, repoData.getRepoTypesList().size());
            Assert.assertNull(repoData.getRepoList());
        } else {
            Assert.assertEquals(1, repoData.getRepoList().size());
            Assert.assertNull(repoData.getRepoTypesList());
        }
    }

    /*
    @Test
    public void testGetPermissionRepoData() {
        Mockito.when(repoService.getLocalAndCachedRepoDescriptorsByRepo(PROJECT_A))
                .thenReturn(Lists.newArrayList(createLocalRepoDescriptor(LOCAL_PROJ),
                        createLocalRepoDescriptor("pipe-info", RepoType.PipeInfo)));

        List<RepoKeyType> project = getAllRepositoriesService.getPermissionRepoData(PROJECT_A);
    }
    */

    private LocalRepoDescriptor createLocalRepoDescriptor(String repoKey) {
        return createLocalRepoDescriptor(repoKey, RepoType.Alpine);
    }

    private LocalRepoDescriptor createLocalRepoDescriptor(String repoKey, RepoType type) {
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setType(type);
        localRepoDescriptor.setKey(repoKey);
        return localRepoDescriptor;
    }

    @Parameterized.Parameters(name = "testExecuteWithDifferentActions: {0}")
    public static Collection provideAction() {
        return Arrays.asList(new Object[][] {
                {"search", "repoTypesList", 2},
                {"deploy", "repoTypesList", 2},
                {"packageSearch", "repoTypesList", 2},
                {"user", "repoTypesList", 3},
                {"local", "repoList", 2},
                {"permission", "repoTypesList", 2},
                {"backup", "repoList", 2},
        });
    }
}